/* 
 * FILE --- fuzzy_ts_sys.hh
 *
 * PURPOSE --- Header file for a class for a multiple input single output 
 * (MISO) Takagi-Sugeno (TS) fuzzy system. The user specifies the number of inputs to the
 * fuzzy system, the number of the Takagi-Sugeno inputs, the number of membership 
 * functions (MF) on each universe of discourse (UD). The MF can be defined as triangular 
 * with 50% of overlapp or as gaussian. Either minimum or product can be used for 
 * premises whereas the defuzzification method is fixed as singleton defuzzification.
 * The utility has a build-in adaptation capability of the output parameters (TS coefficients)
 * based on the Normalized Gradient Method or the so called Projection Algorithm.
 * Therefore it can be used in both direct and indirect adaptive control schemes as
 * a function approximator.
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 * Uses two other utilities:
 *                  FUZZY_UD --- Fuzzy universe of discourse
 *                  FUZZY_MF --- Fuzzy membership function
 *                  MATRIX --- Fuzzy membership function
 *
 */

// Prevent Multiple Inclusion
#ifndef FUZZY_TS_SYS_HH
#define FUZZY_TS_SYS_HH


// Include Files
#include <stdlib.h>
#include "fuzzy_ud.hh"
#include "matrix.hh"


// Local Definitions and Macros
#define FUZZY_TS_SYS_NO_ERROR 0
#define FUZZY_TS_SYS_MEM_ERROR 1

#define DEFAULT_NUM_INPUTS 2 // Default # of input universes of discourse
#define DEFAULT_NUM_OUTPUTS 1 // Default # of output universes of discourse

#define DEFAULT_NUM_IN_MF 7 // Default # of MF on each input universe 
#define DEFAULT_NUM_OUT_MF 7 // Default # of MF on the output universe 

#define MINIMUM_PREMISE 0   // Use minimum rule for evaluating premises
#define PRODUCT_PREMISE 1   // Use product rule for evaluating premises

#define DEFAULT_MF TRIANGULAR_MF  // Triangular MF is default
#define DEFAULT_PREMISE MINIMUM_PREMISE // Minimum is default for premise
#define DEFAULT_DEFUZZ COG_DEFUZZ  // COG defuzzification is default

#define PI 3.14159265359


/* 
   Define the generic multiple input multiple output (MIMO) fuzzy controller
   class. 
*/
class FUZZY_TS_SYS
{
public:
  
  /* Constructor:
     This is the constructor for the class. It allocates all the needed 
     memory and performs the required initialization for the fuzzy system.
     Parameters:
     _num_of_inputs = The # of inputs to the system
     _num_of_ts_inputs = The # of Takagi-Sugeno inputs to the rule concequences
     _num_of_in_mfs = an array of size _num_of_inputs specifying
     the # of MF on each input UD. If NULL the # of input MF on all 
     input UD is set to DEFAULT_NUM_IN_MF
  */
  FUZZY_TS_SYS(int _num_of_inputs, 
	       int _num_of_ts_inputs,
	       int *_num_of_in_mfs = NULL); 
  
  // Destructor
  ~FUZZY_TS_SYS();
  
  /* Set the gains for each input UD
     _in_gains = an array of size equal to the # of inputs 
     if NULL then all the gains are set to zero         
  */
  void set_input_gains(double *_in_gains = NULL);
  
  /* Sets the rule for AND operationin the premise.
     _premise_type = The rule used for the AND operation in the premises
     of the rules. Can be either MINIMUM_PREMISE or PRODUCT_PREMISE. The 
     default is MINIMUM_PREMISE.
  */
  void set_premise_type(int _premise_type = MINIMUM_PREMISE);
  
  // Initialize the TS coefficients
  void set_ts_coeff(const double **_ts_coeff);
  void set_ts_coeff(const double *_ts_coeff);
  void set_ts_coeff(const double _ts_coeff);
  void set_ts_coeff(double _low, double _high);
 
  // Functions to set and return the adaptation gain
  void set_eta(double _eta) { eta = _eta; };
  double get_eta(void) { return eta; };
  
  // Function to return the norm-square of the regressor
  double get_phi_norm(void) { return phi_norm; };

  // The function that given the error updates the Takagi-Sugeno
  // coefficients (i.e., the output parameters) 
  void update_ts_coeff(double error);
  void update_ts_coeff(double error, double norm2);
  void update_ts_coeff(double error, double norm2, double uu);
  void bound_ts_coeff(double _low, double _high);

  /* 
     The function that implements the controller: given an array of crisp 
     inputs to the premices of the rules and the TS inputs to the utputs of the
     rules evaluates the crisp output of the TS fuzzy controller
     
     _input = an array of size equal to the # of inputs providing
     the current inputs to the premices of the controller
     _ts_input = an array of size equal to the # of TS inputs providing
     the current TS inputs to the rule outputs of the controller
  */
  double control(const double *_input, const double *_ts_input);
  
  // Performs a defuzzyfication and returns a crisp output.
  double defuzzyfy();
  
  // Get the private members in case we need it
  int get_premise_type(void) { return premise_type; };
  int get_num_of_inputs(void) { return num_of_inputs; };
  int get_num_of_ts_inputs(void) { return num_of_ts_inputs; };
  int get_num_of_rules(void) { return num_of_rules; };
  
  fuzzy_ud **in_ud;   // An array of input UD
  
  matrix *ts_coeff;  // Coefficients for the ts functions
  
  int fuzzy_ts_sys_err; // Variable to hold the error condition
  
private:
  void allocate_memory(int *_num_of_in_mfs);
  void free_memory(void);

  // Fuzzyfy the given input
  inline void fuzzyfy(const double *_input);

  // Transfere the input to a matrix for matrix operations
  inline void set_ts_input_vector(const double *_ts_input);

  // Calculate the regression vector
  inline void get_phi(void);

  // Functions to transfer parameters (used only if parameters are 
  // updated)
  inline void A2theta(void);
  inline void theta2A(void);
  
  int num_of_inputs;  // # of inputs
  int num_of_ts_inputs;  // # of Takagi-Sugeno inputs 
  int num_of_rules; // # of the rules
  int num_of_out_params; // # of the output parametrs

  // Conveniance variables to be used during calculation of the 
  // controller output
  matrix *ts_inputs;
  matrix *inference;
  
  // Type of rule for evaluating premises. 
  // Can be either MINIMUM_PREMISE or PRODUCT_PREMISE.
  int premise_type; 

  // The the vectros for the parametrs
  matrix *theta;     // An array of output parameters
  matrix *phi;       // An array of function values of the neurons

  double phi_norm; // Variable to hold the norm of phi 
  double eta; // Adaptation gain

  double total_inference; // Sum of the rule inferences 
};


#endif // FUZZY_TS_SYS_HH
