/* 
 * FILE --- fuzzy_ud.hh
 *
 * PURPOSE --- Header file for a class for fuzzy universe of discourse (UD). 
 * The membership functions (MF) can be chosen to be normalized 
 * (i.e. b/n [-1, 1]) with equal spreads and 1/2 overlapp, or can be 
 * chosen b/n user defined range, [_first_center, _last_center] with user
 * defined equal spreads. In these options all the MF are the same user 
 * defined type. The possible MF types are triangular and Gaussian. 
 * Another option would be to initialize each individual MF with its own 
 * center, spread and type. 
 *
 * Uses theutility:
 *              FUZZY_MF --- Fuzzy membership function
 *
 */

// Prevent Multiple Inclusion
#ifndef FUZZY_UD_HH
#define FUZZY_UD_HH

// Include Files
#include <stdlib.h>
#include "fuzzy_mf.hh"

// Local Definitions and Macros
#define FUZZY_UD_NO_ERROR 0
#define FUZZY_UD_MEM_ERROR 1

#define DEFAULT_NUM_MF 7 // Default # of MF on each input UD


class FUZZY_UD
{
public:
  FUZZY_UD(int _num_of_mfs = DEFAULT_NUM_MF);

  ~FUZZY_UD();

  // Initialize the MF
  // Normalized triangular MF b/n [-1, 1] with equal spreads 
  // if _spread is 0 then the MF are initialized with spreads with 1/2 overlapp
  void init_ud(int _mf_type = TRIANGULAR_MF, double _spread = 0);

  // MF centers b/n [_first_center, _last_center] with spreads 
  // specified by _spread. If _spread is 0 then the MF are initalized 
  // with spreads with 1/2 overlapp
  void init_ud(double _first_center, double _last_center,
	       int _mf_type, double _spread = 0);

  // _centers = an array of size num_of_mfs specifying the MF centers
  // _spreads = an array of size num_of_mfs specifying the MF spreads
  // _mf_types = an array of size num_of_mfs specifying the MF types
  void init_ud(double *_centers, double *_spreads, int *_mf_types);

  // _mfs = an array of size num_of_mfs specifying the MF
  void init_ud(fuzzy_mf *_mfs);

  // Return the # of MF on this universe
  int get_num_of_mfs(void) const { return num_of_mfs; }

  // Fuzzyfies the given input on this UD. In other words, evaluates 
  // the degree of membership for all MF.
  void fuzzyfy(const double _input);

  double ud_gain;    // Gain for this UD
  double *deg_of_mbrship;  // Degree of membeship for each MF
  int fuzzy_ud_err;     // Error variable
  fuzzy_mf *mfs;        // An array of MF

private:
  void allocate_memory(void);
  void free_memory(void);

  int num_of_mfs; // # of MF in this UD

  double scaled_input; // Conveniance variable
};

typedef FUZZY_UD fuzzy_ud;

#endif // FUZZY_UD_HH
