 /*
  * FILE --- veh_sim_fuzzy_ts.cc
  *
  *
  * PURPOSE --- To provide an example use of the class for 4'th
  * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
  * n-input 1-output Takagi-Sugeno (TS) fuzzy system, FUZZY_TS_SYS.
  *
  * Implements the member functions of a class for performing a simulation of
  * a simple vehicle with point mass dynamics controlled by a TS fuzzy controller.
  * The fuzzy controller is implemented as a part of the class for the vehicle
  * dynamics - VEH_SIM_FUZZY_TS. 
  *
  */

// Include Files
#include "veh_sim_fuzzy_ts.hh"

VEH_SIM_FUZZY_TS::VEH_SIM_FUZZY_TS (double _sampling_period, int _simulation_per_cycle,
				    double _initial_time, double *_initial_state) 
  : RK4_SIM(3, 3, (_sampling_period/((double)_simulation_per_cycle)), 
	    _initial_time, _initial_state)
{
  // samp_period >= integration_step
  
  // Initialize parameters
  sampling_period = _sampling_period;
  simulation_per_cycle = _simulation_per_cycle;
  mass = 1300;
  drag_x = 0.3;
  //  drag_y = 0.1;
  tau = 0.2;
  disturbance = 100.0;

  output_gain = 0;
  
  // TS Fuzzy controller - SISO
  fuzzy_ts_sys = new FUZZY_TS_SYS(1, 0);
}

VEH_SIM_FUZZY_TS::~VEH_SIM_FUZZY_TS()
{
  delete fuzzy_ts_sys;
}

inline void VEH_SIM_FUZZY_TS::DERIV(const double t, const double *x, 
				    const double *u, double *xdot)
{
  // This is where the derivative is defined: xdot = f(t,x,u)
  // Simple point mass dynamics
  // delta_dot = V_veh - V_preceading;
  // V_veh_dot = (1.0/m)*(-Ap*(V_veh)^2 - d + f);
  // f_dot = (1.0/tau)*(-f + u);
  
  // x[0] = longitudinal position
  // x[1] = longitudinal velosity
  // x[2] = force in longitudinal direction
  // u[0] = reference velocity
  // u[1] = applied control input
  
  // Vehicle Dynamics
  xdot[0] = x[1] - u[0];
  xdot[1] = (1.0/mass)*(-drag_x*x[1]*x[1] - disturbance + x[2]);
  xdot[2] = (1.0/tau)*(-x[2] + u[1]);
}

// Each sampling time the controller is called in this function and
// based on the current control input the vehicle dynamics are simulated
void VEH_SIM_FUZZY_TS::simulate()
{
  static double ref_vel = 45.0;
  static double lambda = 0.9;
  double err;
  static int count = 1000;

  // The objective is to drive this error to zero.
  // The fuzzy controller is SISO but I defined the simulator
  // as there were 3 controller inputs. I use two of them to 
  // save the error for plotting purposes.
  err = (plant_state[0] + lambda*plant_state[1]);
  
  // I use this for the refeference velocity. In real situation
  // This will be the velocity of the preceading vehicle and
  // will be measured.
  ctrl_input[0] = ref_vel;

  // Actual control input is ctrl_input[1]
  ctrl_input[1] = - output_gain * fuzzy_ts_sys->control(&err, plant_state);

  // Simulate one sampling period
  for (int ii = 0; ii < simulation_per_cycle; ii++)
    {
      solve();
    }
  
  // I use this just for saving the error
  ctrl_input[2] = err;
  
  // Trapezoidal reference input --- the front driver is creazy :)
  if(count < 2000)
    ref_vel += 0.005;
  else if((count >= 3000) && (count < 5000))
    ref_vel -= 0.005;
  count++;
  if(count == 6000)
    count = 0;
}


